/**
 * Copyright (c) 2012 - 2017 Data In Motion and others.
 * All rights reserved. 
 * 
 * This program and the accompanying materials are made available under the terms of the 
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Data In Motion - initial API and implementation
 */
package org.eclipse.ece.messaging;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import org.eclipse.ece.messaging.core.BasicExampleTest;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.gecko.osgi.messaging.Message;
import org.gecko.osgi.messaging.MessagingConstants;
import org.gecko.osgi.messaging.MessagingService;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.runners.MockitoJUnitRunner;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Filter;
import org.osgi.framework.FrameworkUtil;
import org.osgi.service.cm.Configuration;
import org.osgi.util.pushstream.PushStream;

/**
 * Test that shows simple mapping and filtering
 * @author Mark Hoffmann
 * @since 18.10.2017
 */
@RunWith(MockitoJUnitRunner.class)
public class Example01FilterAndMapTest extends BasicExampleTest {

	private final BundleContext context = FrameworkUtil.getBundle(Example01FilterAndMapTest.class).getBundleContext();
	private Configuration mqttConfig = null;
	private Configuration wsConfig = null;
	
	@Before
	public void before() throws MqttException {
		setup(context);
	}

	@After
	public void after() throws MqttException, IOException {
		teardown();
		if (wsConfig != null) {
			wsConfig.delete();
			wsConfig = null;
		}
		if (mqttConfig != null) {
			mqttConfig.delete();
			mqttConfig = null;
		}
	}

	@Test
	public void testFilterAndMap() throws Exception {
		
		mqttConfig = getConfiguration("MQTTService");
		assertNotNull(mqttConfig);
		
		Dictionary<String, Object> p = mqttConfig.getProperties();
		assertNull(p);
		
		p = new Hashtable<>();
		p.put(MessagingConstants.PROP_BROKER, MQTT_BROKER_URL);
		p.put("type", "mqtt");

		mqttConfig.update(p);

		// check for service
		Filter mqttFilter = FrameworkUtil.createFilter("(type=mqtt)");
		MessagingService messagingService = getService(mqttFilter, 30000l);
		assertNotNull(messagingService);

		//send message and start automatic publishing
		messagingService.publish(MQTT_TOPIC_TEMP, ByteBuffer.wrap("start c-mqtt01".getBytes()));
		
		PushStream<Message> subscribe01 = messagingService.subscribe(MQTT_TOPIC_TEMP);
		subscribe01.forEach((m)->System.out.println("Push stream 01: " + new String(m.payload().array())));
		
		PushStream<Message> subscribe02 = messagingService.subscribe(MQTT_TOPIC_TEMP);
		subscribe02.map(this::convertMessageToString).filter((str)->str.contains("mqtt02")).forEach((str)->System.out.println("Push stream 02 with filtering and mapping: " + str));
		
//		PushStream<Message> merge = subscribe01.merge(subscribe02);
//		merge.map(this::convertMessageToString).forEach((str)->System.out.println("Client 01 and 02: " + str));
//		merge.distinct().map(this::convertMessageToString).forEach((str)->System.out.println("Client 01 and 02: " + str));
		
		CountDownLatch waitLatch = new CountDownLatch(1);
		waitLatch.await(10, TimeUnit.SECONDS);
		
		// start second client
		messagingService.publish(MQTT_TOPIC_TEMP, ByteBuffer.wrap("start c-mqtt02".getBytes()));
		
		waitLatch = new CountDownLatch(1);
		waitLatch.await(10, TimeUnit.SECONDS);
		
		//stop automatic publishing for first client
		messagingService.publish(MQTT_TOPIC_TEMP, ByteBuffer.wrap("stop c-mqtt01".getBytes()));
		
		waitLatch = new CountDownLatch(1);
		waitLatch.await(10, TimeUnit.SECONDS);
		
		// stop automatic publishing for second client
		messagingService.publish(MQTT_TOPIC_TEMP, ByteBuffer.wrap("stop c-mqtt02".getBytes()));
	}
	
	/**
	 * Converts the payload into a {@link String}
	 * @param message the {@link Message}
	 * @return the message string
	 */
	private String convertMessageToString(Message message) {
		return new String(message.payload().array());
	}
	
}