/**
 * Copyright (c) 2012 - 2017 Data In Motion and others.
 * All rights reserved. 
 * 
 * This program and the accompanying materials are made available under the terms of the 
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Data In Motion - initial API and implementation
 */
package org.eclipse.ece.messaging;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import org.eclipse.ece.messaging.core.BasicExampleTest;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.gecko.osgi.messaging.Message;
import org.gecko.osgi.messaging.MessagingConstants;
import org.gecko.osgi.messaging.MessagingService;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.runners.MockitoJUnitRunner;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Filter;
import org.osgi.framework.FrameworkUtil;
import org.osgi.service.cm.Configuration;
import org.osgi.util.pushstream.PushStream;

/**
 * Shows message adapter with MQTT and Websockets
 * @author Mark Hoffmann
 * @since 18.10.2017
 */
@RunWith(MockitoJUnitRunner.class)
public class Example02MQTTAndWebsocketsTest extends BasicExampleTest {

	private final BundleContext context = FrameworkUtil.getBundle(Example02MQTTAndWebsocketsTest.class).getBundleContext();
	private Configuration mqttConfig = null;
	private Configuration wsConfig = null;
	
	@Before
	public void before() throws MqttException {
		setup(context);
	}

	@After
	public void after() throws MqttException, IOException {
		if (wsConfig != null) {
			wsConfig.delete();
			wsConfig = null;
		}
		if (mqttConfig != null) {
			mqttConfig.delete();
			mqttConfig = null;
		}
	}

	@Test
	public void testPushStreamWebsocketAndMqtt() throws Exception {
		
		mqttConfig = getConfiguration("MQTTService");
		wsConfig = getConfiguration("WSService");

		// has to be a new configuration
		Dictionary<String, Object> mqttProperties = mqttConfig.getProperties();
		assertNull(mqttProperties);
		
		mqttProperties = new Hashtable<>();
		mqttProperties.put(MessagingConstants.PROP_BROKER, MQTT_BROKER_URL);
		mqttProperties.put("type", "mqtt");
		
		// has to be a new configuration
		Dictionary<String, Object> wsProperties = mqttConfig.getProperties();
		assertNull(wsProperties);
		
		wsProperties = new Hashtable<>();
		wsProperties.put(MessagingConstants.PROP_BROKER, WS_BROKER_URL);
		wsProperties.put("type", "ws");

		// starting adapter with the given properties
		wsConfig.update(wsProperties);
		mqttConfig.update(mqttProperties);

		// check for mqtt service
		Filter mqttFilter = FrameworkUtil.createFilter("(type=mqtt)");
		MessagingService mqttService = getService(mqttFilter, 30000l);
		assertNotNull(mqttService);
		
		// check for ws service
		Filter wsFilter = FrameworkUtil.createFilter("(type=ws)");
		MessagingService wsService = getService(wsFilter, 30000l);
		assertNotNull(wsService);
		

		//send message and start publishing
		mqttService.publish(MQTT_TOPIC_TEMP, ByteBuffer.wrap("start c-mqtt01".getBytes()));

		// push stream delivering MQTT temperature data
		PushStream<Message> subscribe01 = mqttService.subscribe(MQTT_TOPIC_TEMP);
		subscribe01.forEach((m)->System.out.println("Push stream 1 MQTT temperatur: " + new String(m.payload().array())));
		
		// push stream delivering MQTT wind data
		PushStream<Message> subscribe02 = mqttService.subscribe(MQTT_TOPIC_WIND);
		subscribe02.map(this::convertMessageToString).filter((str)->str.contains("mqtt03")).forEach((str)->System.out.println("Push stream 2 MQTT wind: " + str));
	
		// push stream delivering Websocket data
		PushStream<Message> subscribe03 = wsService.subscribe(WS_PATH);
		subscribe03.map(this::convertMessageToString).filter((str)->str.contains("ws-server")).forEach((str)->System.out.println("Push stream 3 WS: " + str));
		
		CountDownLatch waitLatch = new CountDownLatch(1);
		waitLatch.await(10, TimeUnit.SECONDS);
		
		// start mqtt publishing for wind
		mqttService.publish(MQTT_TOPIC_WIND, ByteBuffer.wrap("start c-mqtt03".getBytes()));

		waitLatch = new CountDownLatch(1);
		waitLatch.await(10, TimeUnit.SECONDS);
		
		//send ws message to start automatic publishing
		wsService.publish(WS_PATH, ByteBuffer.wrap("start".getBytes()));
		
		waitLatch = new CountDownLatch(1);
		waitLatch.await(5, TimeUnit.SECONDS);
		
		//stop all automatic publishing
		mqttService.publish(MQTT_TOPIC_TEMP, ByteBuffer.wrap("stop c-mqtt01".getBytes()));
		mqttService.publish(MQTT_TOPIC_WIND, ByteBuffer.wrap("stop c-mqtt03".getBytes()));
		wsService.publish(WS_PATH, ByteBuffer.wrap("stop".getBytes()));
	}
	
	/**
	 * Converts the payload into a {@link String}
	 * @param message the {@link Message}
	 * @return the message string
	 */
	private String convertMessageToString(Message message) {
		return new String(message.payload().array());
	}
	
}