/**
 * Copyright (c) 2012 - 2017 Data In Motion and others.
 * All rights reserved. 
 * 
 * This program and the accompanying materials are made available under the terms of the 
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Data In Motion - initial API and implementation
 */
package org.eclipse.ece.messaging;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;

import org.eclipse.ece.messaging.core.BasicExampleTest;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.gecko.osgi.messaging.Message;
import org.gecko.osgi.messaging.MessagingConstants;
import org.gecko.osgi.messaging.MessagingService;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.runners.MockitoJUnitRunner;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Filter;
import org.osgi.framework.FrameworkUtil;
import org.osgi.service.cm.Configuration;
import org.osgi.util.pushstream.PushStream;

/**
 * Show how back-pressure works in action
 * @author Mark Hoffmann
 * @since 18.10.2017
 */
@RunWith(MockitoJUnitRunner.class)
public class Example05BackPressureTest extends BasicExampleTest {

	private final BundleContext context = FrameworkUtil.getBundle(Example05BackPressureTest.class).getBundleContext();
	private Configuration mqttConfig = null;
	
	@Before
	public void before() throws MqttException {
		setup(context);
	}

	@After
	public void after() throws MqttException, IOException {
		if (mqttConfig != null) {
			mqttConfig.delete();
			mqttConfig = null;
		}
	}

	@Test
	public void testPushstreamBackpressure() throws Exception {
		
		mqttConfig = getConfiguration("MQTTService");

		// has to be a new configuration
		Dictionary<String, Object> mqttProperties = mqttConfig.getProperties();
		assertNull(mqttProperties);
		
		mqttProperties = new Hashtable<>();
		mqttProperties.put(MessagingConstants.PROP_BROKER, MQTT_BROKER_URL);
		mqttProperties.put("type", "mqtt");
		
		mqttConfig.update(mqttProperties);

		// check for mqtt service
		Filter mqttFilter = FrameworkUtil.createFilter("(type=mqtt)");
		MessagingService mqttService = getService(mqttFilter, 30000l);
		assertNotNull(mqttService);
		AtomicLong al1 = new AtomicLong(System.currentTimeMillis());
		AtomicLong al2 = new AtomicLong(System.currentTimeMillis());
		
		// start automatic publishing
		mqttService.publish(MQTT_TOPIC_TEMP, ByteBuffer.wrap("start c-mqtt01".getBytes()));

		// Push stream one using back pressure and buffering
		PushStream<Message> subscribe01 = mqttService.subscribe(MQTT_TOPIC_TEMP);
		subscribe01.adjustBackPressure((m)->{return 1500l;}).map(this::convertMessageToString).forEach((s)->{
			long currentTime = System.currentTimeMillis();
			long oldTime = al1.getAndSet(currentTime);
			long time = currentTime - oldTime;
			System.out.println("Push stream 1 back-pressure: '" + s + "' took " + time + "ms");
		});
		
		// Push stream without back pressure configuration
		PushStream<Message> subscribe02 = mqttService.subscribe(MQTT_TOPIC_TEMP);
		subscribe02.map(this::convertMessageToString).forEach((s)->{
			long currentTime = System.currentTimeMillis();
			long oldTime = al2.getAndSet(currentTime);
			long time = currentTime - oldTime;
			System.out.println("Push stream 2 no back-pressure: '" + s + "' took " + time + "ms");
		});
		
		CountDownLatch waitLatch = new CountDownLatch(1);
		waitLatch.await(10, TimeUnit.SECONDS);
		
		mqttService.publish(MQTT_TOPIC_TEMP, ByteBuffer.wrap("stop c-mqtt01".getBytes()));
		System.out.println("Stop producing data for both push streams");
		
		waitLatch = new CountDownLatch(1);
		waitLatch.await(10, TimeUnit.SECONDS);

	}
	
	/**
	 * Converts the payload into a {@link String}
	 * @param message the {@link Message}
	 * @return the message string
	 */
	private String convertMessageToString(Message message) {
		return new String(message.payload().array());
	}
	
}